/*
   orxonox - the future of 3D-vertical-scrollers

   Copyright (C) 2004 orx

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   ### File Specific:
   main-programmer: Benjamin Grauer
   co-programmer: ...
*/

#define DEBUG_SPECIAL_MODULE DEBUG_MODULE_SHELL

#include "shell_buffer.h"

#include <stdarg.h>

#include "debug.h"
#include "shell.h"
#include "lib/util/threading.h"

namespace OrxShell
{
  /**
   * @brief standard constructor
   */
  ShellBuffer::ShellBuffer ()
  {
    ShellBuffer::singletonRef = this;
    this->shell = NULL;

    this->lineCount = 0;
    this->bufferArray[0] = '\0';

    this->setBufferSize(100);
  }

  ShellBuffer* ShellBuffer::singletonRef = NULL;
  std::list<std::string> ShellBuffer::buffer;

  /**
   * @brief standard deconstructor
   */
  ShellBuffer::~ShellBuffer ()
  {
    if (this->shell != NULL)
      delete this->shell;

    ShellBuffer::singletonRef = NULL;
  }

  /**
   * @brief registers the Shell to the Buffer
   * @param shell the Shell to register.
   */
  void ShellBuffer::registerShell(Shell* shell)
  {
    if (this->shell == NULL)
      this->shell = shell;
    else
      PRINTF(1)("already registered a Shell to the ShellBuffer\n");
  }

  /**
   * @brief unregisters the Shell from the Buffer
   * @param shell the Shell to unregister.
   */
  void ShellBuffer::unregisterShell(Shell* shell)
  {
    if (this->shell == shell)
      this->shell = NULL;
    else
      PRINTF(1)("cannot unregister shell, because it is not registered to the ShellBuffer\n");
  }

  /**
   * @brief deletes all the Buffers
   */
  void ShellBuffer::flush()
  {
    this->buffer.clear();
  }

  /**
   * @brief adds a new Line to the List of Buffers
   * @param line the Line as in the first argument in printf
   */
  bool ShellBuffer::addBufferLineStatic(const char* line, ...)
  {
    va_list arguments;
    va_start(arguments, line);

    static OrxThread::Mutex ShellBuffer__bufferMutex;

    OrxThread::MutexLock bufferLock(&ShellBuffer__bufferMutex);
#if DEBUG_LEVEL < 3
    if (ShellBuffer::singletonRef == NULL)
#endif
      vprintf(line, arguments);
#if DEBUG_LEVEL < 3
    else
#else
    if (ShellBuffer::singletonRef != NULL)
#endif
      ShellBuffer::singletonRef->addBufferLine(line, arguments);
    return true;
  }

  /**
   * @brief add a Line to the List of Buffers
   * @param line
   * @param arguments
   *
   * This function Adds one line to the buffer.
   * and displays the line as the First Line of the display-buffer
   */
  void ShellBuffer::addBufferLine(const char* line, va_list arguments)
  {
    // copy the output to the bufferArray
    vsprintf(this->bufferArray, line, arguments);

    std::string inputBuffer = this->keepBuffer + this->bufferArray;

    int lineBegin = 0;
    int lineEnd = 0;
    // adding all the new Lines
    while (lineEnd < inputBuffer.size())
    {
      lineBegin = lineEnd;
      lineEnd = inputBuffer.find('\n', (lineBegin == 0) ? 0: ++lineBegin);
      if (likely(lineEnd != std::string::npos ))
      {
        this->lineCount++;

        printf("Test:: %s \n", inputBuffer.substr(lineBegin, lineEnd - lineBegin).c_str());
        this->buffer.push_front(inputBuffer.substr(lineBegin, lineEnd - lineBegin));
        if (likely (this->shell != NULL) && unlikely (this->shell->isActive()))
          this->shell->printToDisplayBuffer(this->buffer.front());
      }
      else      // No end of Line reached.
      {
        this->keepBuffer = inputBuffer.substr(lineBegin, inputBuffer.size() - lineBegin);
        break;
      }



      if (this->buffer.size() > this->bufferSize)
        this->buffer.pop_back();
    }
  }

  /**
   * @brief displays some nice output from the Shell
   */
  void ShellBuffer::debug() const
  {
    PRINT(3)("Debugging output to console (not this shell)\n");

    std::list<std::string>::const_iterator bufferLine;
    for (bufferLine = --this->buffer.end(); bufferLine != this->buffer.begin(); --bufferLine)
      printf((*bufferLine).c_str());
  }

}
